/*
 * Copyright (c) 2012 - 2018, Nordic Semiconductor ASA
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 *
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**@example examples/heart_rate_monitor
 *
 * @brief Heart Rate Service Sample Application main file.
 *
 * This file contains the source code for a sample application using the Heart Rate service.
 * This service exposes heart rate data from a Heart Rate Sensor intended for fitness applications.
 * https://www.bluetooth.com/specifications/gatt/viewer?attributeXmlFile=org.bluetooth.service.heart_rate.xml
 *
 * Structure of this file
 * - Includes
 * - Definitions
 * - Global variables
 * - Global functions
 * - Event functions
 * - Event dispatcher
 * - Main
 */

/** Includes */
#ifndef _WIN32
#define _DEFAULT_SOURCE // Get rid of usleep warnings
#endif

#include "ble.h"
#include "sd_rpc.h"

#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stddef.h>

#include "utils.h"
#include "BTLE_Assigned_Numbers.h"
#include "HeartRateControlStructs.h"

#ifdef _WIN32
#include <windows.h>
#else
#include <unistd.h>
#define Sleep(x) usleep((x)*1000)
#endif


/** Definitions */
#define DEFAULT_BAUD_RATE 1000000 /**< The baud rate to be used for serial communication with nRF5 device. */

#ifdef _WIN32
#define DEFAULT_UART_PORT_NAME "COM5"
#endif
#ifdef __APPLE__
#define DEFAULT_UART_PORT_NAME "/dev/tty.usbmodem00000"
#endif
#ifdef __linux__
#define DEFAULT_UART_PORT_NAME "/dev/ttyACM0"
#endif

//#define ADVERTISING_INTERVAL_40_MS 64  /**< 0.625 ms = 40 ms */
#define ADVERTISING_TIMEOUT_3_MIN  180 /**< 1 sec = 3 min */

#define OPCODE_LENGTH 1                                                 /**< Length of opcode inside Heart Rate Measurement packet. */
#define HANDLE_LENGTH 2                                                 /**< Length of handle inside Heart Rate Measurement packet. */

#define MAX_HRM_LEN (BLE_EVT_LEN_MAX(GATT_MTU_SIZE_DEFAULT) - OPCODE_LENGTH - HANDLE_LENGTH)

#define HEART_RATE_BASE     65
#define HEART_RATE_INCREASE 3
#define HEART_RATE_LIMIT    190

//#define BUFFER_SIZE 30           /**< Sufficiently large buffer for the advertising data.  */
//#define DEVICE_NAME "Nordic_HRM" /**< Name device advertises as over Bluetooth. */

#ifndef GATT_MTU_SIZE_DEFAULT
#define GATT_MTU_SIZE_DEFAULT BLE_GATT_ATT_MTU_DEFAULT
#endif


/** Global variables */
static uint16_t                 m_connection_handle             = BLE_CONN_HANDLE_INVALID;
static uint16_t                 m_device_information_service_handle = BLE_GATT_HANDLE_INVALID;
static uint16_t                 m_heart_rate_service_handle     = BLE_GATT_HANDLE_INVALID;
static ble_gatts_char_handles_t m_heart_rate_measurement_handle;
static ble_gatts_char_handles_t m_sensor_location_handle;
static uint8_t                  m_heart_rate                    = HEART_RATE_BASE;
static bool                     m_send_notifications            = false;
static bool                     cccdSet                         = false;
static bool                     m_is_encrytped					= false;
static bool                     m_advertisement_timed_out       = false;
static adapter_t *              m_adapter                       = NULL;

static uint16_t                 m_dis_service_handle            = BLE_GATT_HANDLE_INVALID;
static ble_gatts_char_handles_t m_dis_system_id_handle;


static uint32_t                 m_config_id                     = 1;

static uint8_t                  m_adv_handle                    = 0;
//static ble_gap_adv_params_t     m_adv_params;

static unsigned long long elapsedTime;
static unsigned long long elapsedTimeStart;

volatile static bool pairResponse = false;
ble_gap_sec_params_t* own_sec_params = NULL;
ble_gap_sec_keyset_t* keys = NULL;
uint16_t saveDataLength = 0;
uint8_t *saveDataBuffer = NULL;
bool hasPaired = false;
s_hrParams* hrParams;


/** Global functions */

void setGlobals()
{
    m_connection_handle = BLE_CONN_HANDLE_INVALID;
    m_device_information_service_handle = BLE_GATT_HANDLE_INVALID;
    m_heart_rate_service_handle = BLE_GATT_HANDLE_INVALID;
    memset(&m_heart_rate_measurement_handle, 0, sizeof(m_heart_rate_measurement_handle));
    memset(&m_sensor_location_handle, 0, sizeof(m_sensor_location_handle));
    m_heart_rate = HEART_RATE_BASE;
    m_send_notifications = false;
    cccdSet = false;
    m_is_encrytped = false;
    m_advertisement_timed_out = false;
    m_adapter = NULL;

    m_dis_service_handle = BLE_GATT_HANDLE_INVALID;
    memset(&m_dis_system_id_handle, 0, sizeof(m_dis_system_id_handle));


    m_config_id = 1;

    m_adv_handle = 0;
    

    elapsedTime = 0;
    elapsedTimeStart = 0;

    pairResponse = false;
    own_sec_params = NULL;
    keys = NULL;
    saveDataLength = 0;
    saveDataBuffer = NULL;
    hasPaired = false;
    hrParams = NULL;
}

/**@brief Function for handling error message events from sd_rpc.
 *
 * @param[in] adapter The transport adapter.
 * @param[in] code Error code that the error message is associated with.
 * @param[in] message The error message that the callback is associated with.
 */
static void status_handler(adapter_t * adapter, sd_rpc_app_status_t code, const char * message)
{
    printf("Status: %d, message: %s\n", (uint32_t)code, message);
    fflush(stdout);
}

/**@brief Function for handling the log message events from sd_rpc.
 *
 * @param[in] adapter The transport adapter.
 * @param[in] severity Level of severity that the log message is associated with.
 * @param[in] message The log message that the callback is associated with.
 */
static void log_handler(adapter_t * adapter, sd_rpc_log_severity_t severity, const char * message)
{
    switch (severity)
    {
        case SD_RPC_LOG_ERROR:
            printf("Error: %s\n", message);
            fflush(stdout);
            break;

        case SD_RPC_LOG_WARNING:
            printf("Warning: %s\n", message);
            fflush(stdout);
            break;

        case SD_RPC_LOG_INFO:
            printf("Info: %s\n", message);
            fflush(stdout);
            break;

        default:
            printf("Log: %s\n", message);
            fflush(stdout);
            break;
    }
}

/**@brief Function for initializing serial communication with the target nRF5 Bluetooth slave.
 *
 * @param[in] serial_port The serial port the target nRF5 device is connected to.
 *
 * @return The new transport adapter.
 */
static adapter_t * adapter_init(char * serial_port, uint32_t baud_rate)
{
    physical_layer_t  * phy;
    data_link_layer_t * data_link_layer;
    transport_layer_t * transport_layer;

    phy = sd_rpc_physical_layer_create_uart(serial_port,
                                            baud_rate,
                                            SD_RPC_FLOW_CONTROL_NONE,
                                            SD_RPC_PARITY_NONE);
    data_link_layer = sd_rpc_data_link_layer_create_bt_three_wire(phy, 250);
    transport_layer = sd_rpc_transport_layer_create(data_link_layer, 1500);
    return sd_rpc_adapter_create(transport_layer);
}

/**@brief Function for initializing the BLE stack.
 *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t ble_stack_init()
{
    uint32_t            err_code;
    uint32_t *          app_ram_base = NULL;


    err_code = sd_ble_enable(m_adapter, app_ram_base);

    switch (err_code)
	{
		case NRF_SUCCESS:
			break;
		case NRF_ERROR_INVALID_STATE:
printf("BLE stack already enabled\n");
fflush(stdout);
break;
		default:
            printf("Failed to enable BLE stack. Error code: %d\n", err_code);
            fflush(stdout);
            break;
    }

    return err_code;
}


/**@brief Function for setting configuration for the BLE stack.
 *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
uint32_t ble_cfg_set(uint8_t conn_cfg_tag)
{
    const uint32_t ram_start = 0; // Value is not used by ble-driver
    uint32_t error_code;
    ble_cfg_t ble_cfg;

    // Configure the connection roles.
    memset(&ble_cfg, 0, sizeof(ble_cfg));
    ble_cfg.gap_cfg.role_count_cfg.periph_role_count = 1; // maximum number of connections as a peripheral
    ble_cfg.gap_cfg.role_count_cfg.central_role_count = 1;
    ble_cfg.gap_cfg.role_count_cfg.central_sec_count = 1;

    error_code = sd_ble_cfg_set(m_adapter, BLE_GAP_CFG_ROLE_COUNT, &ble_cfg, ram_start);
    if (error_code != NRF_SUCCESS)
    {
        printf("sd_ble_cfg_set() failed when attempting to set BLE_GAP_CFG_ROLE_COUNT. Error code: 0x%02X\n", error_code);
        fflush(stdout);
        return error_code;
    }

    memset(&ble_cfg, 0x00, sizeof(ble_cfg));
    ble_cfg.conn_cfg.conn_cfg_tag = conn_cfg_tag;
    ble_cfg.conn_cfg.params.gatt_conn_cfg.att_mtu = 150;

    error_code = sd_ble_cfg_set(m_adapter, BLE_CONN_CFG_GATT, &ble_cfg, ram_start);
    if (error_code != NRF_SUCCESS)
    {
        printf("sd_ble_cfg_set() failed when attempting to set BLE_CONN_CFG_GATT. Error code: 0x%02X\n", error_code);
        fflush(stdout);
        return error_code;
    }

    return NRF_SUCCESS;
}



/**@brief Function for initializing the Advertising functionality and starting advertising.
 *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t advertising_start()
{
    uint32_t             error_code;
    ble_gap_adv_params_t adv_params;
    memset(&adv_params, 0, sizeof(adv_params));

    error_code = sd_ble_gap_adv_start(m_adapter, m_adv_handle, m_config_id);

    if (error_code != NRF_SUCCESS)
    {
        printf("Failed to start advertising. Error code: 0x%02X\n", error_code);
        fflush(stdout);
        return error_code;
    }

    elapsedTimeStart = GetTickCount64();
    printf("Started advertising at time 0\n");
    fflush(stdout);
    return NRF_SUCCESS;
}

/**@brief Function for encoding a Heart Rate Measurement.
 *
 * @param[in]   hrMsmtChar    Measurement to be encoded.
 * @param[out]  encoded_hrm   Buffer where the encoded data will be written.
 *
 * @return      Size of encoded data.
 */
static uint8_t heart_rate_measurement_encode(uint8_t* encoded_hrm, s_hrMsmtChar* hrMsmtChar)
{
    int i;
    int index = 0;

    encoded_hrm[index++] = (hrMsmtChar->energyExpendedStatusFlag | hrMsmtChar->hrFormatFlag |
                                hrMsmtChar->hrSensorContactFlag | hrMsmtChar->RRntervalPresentFlag);
    encoded_hrm[index++] = (hrMsmtChar->heartRate & 0xFF);
    if (hrMsmtChar->hrFormatFlag != 0)
    {
        encoded_hrm[index++] = ((hrMsmtChar->heartRate & 0xFF00) >> 8);
    }
    if (hrMsmtChar->energyExpendedStatusFlag != 0)
    {
        encoded_hrm[index++] = (hrMsmtChar->energyExpended & 0xFF);
        encoded_hrm[index++] = ((hrMsmtChar->energyExpended & 0xFF00) >> 8);
    }
    for (i = 0; i < hrMsmtChar->numberOfRrIntervals; i++)
    {
        encoded_hrm[index++] = (hrMsmtChar->rrIntervals[i] & 0xFF);
        if (index >= MAX_HRM_LEN)
        {
            index--;
            break;
        }
        encoded_hrm[index++] = ((hrMsmtChar->rrIntervals[i] & 0xFF00) >> 8);
        if (index >= MAX_HRM_LEN)
        {
            break;
        }
    }

    return index;
}




/**@brief Function for simulating a heart rate sensor reading.
 *
 * @details This modifies a global variable `m_heart_rate`.
 */
static void heart_rate_generate()
{
    m_heart_rate += HEART_RATE_INCREASE;

    if (m_heart_rate > HEART_RATE_LIMIT)
    {
        m_heart_rate = HEART_RATE_BASE;
    }
}

/**@brief Function for sending the heart rate measurement over Bluetooth.
 *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t heart_rate_measurement_send(s_hrMsmtChar *hrMsmtChar)
{
    uint32_t               error_code;
    uint8_t                encoded_hrm[MAX_HRM_LEN];
    uint16_t               hvx_length;
    ble_gatts_hvx_params_t hvx_params;
    uint8_t                length;

 //   heart_rate_generate();

    length = heart_rate_measurement_encode(encoded_hrm, hrMsmtChar);

    hvx_length = length;

    hvx_params.handle   = m_heart_rate_measurement_handle.value_handle;
    hvx_params.type     = BLE_GATT_HVX_NOTIFICATION;
    hvx_params.offset   = 0;
    hvx_params.p_len    = &hvx_length;
    hvx_params.p_data   = encoded_hrm;

    error_code = sd_ble_gatts_hvx(m_adapter, m_connection_handle, &hvx_params);

    if (error_code != NRF_SUCCESS)
    {
        printf("Failed to send heart rate measurement. Error code: 0x%02X\n", error_code);
        fflush(stdout);
        return error_code;
    }

    return NRF_SUCCESS;
}


/** Event functions */


/** Event dispatcher */

/**@brief Function for handling the Application's BLE Stack events.
 *
 * @param[in] adapter The transport adapter.
 * @param[in] p_ble_evt Bluetooth stack event.
 */
// For pairing

static void ble_evt_dispatch(adapter_t * adapter, ble_evt_t * p_ble_evt)
{
    uint32_t              err_code;
	ble_gap_conn_params_t connParams;
	char                  byteString[64];

    if (p_ble_evt == NULL)
    {
        printf("Received an empty BLE event\n");
        fflush(stdout);
        return;
    }

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            m_connection_handle = p_ble_evt->evt.gap_evt.conn_handle;
            printf("Connected at time %llu, connection handle 0x%04X\n", (GetTickCount64() - elapsedTimeStart), m_connection_handle);
            fflush(stdout);
            if (saveDataBuffer != NULL)
            {
                err_code = sd_ble_gatts_sys_attr_set(m_adapter, p_ble_evt->evt.gap_evt.conn_handle, saveDataBuffer,
                    saveDataLength, BLE_GATTS_SYS_ATTR_FLAG_SYS_SRVCS);
                if (err_code != NRF_SUCCESS)
                {
                    printf("Failed updating persistent sys attr info. Error code: 0x%02X\n", err_code);
                    fflush(stdout);
                }
            }
            if (hrParams->invokeServiceChanged)
            {
                uint8_t* p_sys_attr_data;
                uint16_t p_len = 0;
                uint16_t startHandle = 0;
                sd_ble_gatts_sys_attr_get(m_adapter, m_connection_handle, NULL, &p_len, BLE_GATTS_SYS_ATTR_FLAG_SYS_SRVCS);
                p_sys_attr_data = calloc(1, p_len);
                sd_ble_gatts_sys_attr_get(m_adapter, m_connection_handle, p_sys_attr_data, &p_len, BLE_GATTS_SYS_ATTR_FLAG_SYS_SRVCS);
                sd_ble_gatts_sys_attr_set(m_adapter, m_connection_handle, p_sys_attr_data, p_len, BLE_GATTS_SYS_ATTR_FLAG_SYS_SRVCS);
                //startHandle = (uint16_t)(p_sys_attr_data[0] + (p_sys_attr_data[1] << 8));
                while(true)
                {
                    err_code = sd_ble_gatts_service_changed(m_adapter, m_connection_handle, startHandle, 0xFFFF);
                    if (err_code == BLE_ERROR_INVALID_ATTR_HANDLE)
                    {
                        printf("Failed invoking service changed indication. Invalid handle %d\n", startHandle);
                        startHandle++;
                        if (startHandle < 0xFFFF)
                        {
                            continue;
                        }
                    }
                    else if (err_code != BLE_ERROR_GATTS_SYS_ATTR_MISSING)
                    {
                        printf("Failed invoking service changed indication. Service attributes missing\n");
                        fflush(stdout);
                    }
                    else if (err_code != NRF_SUCCESS)
                    {
                        printf("Failed invoking service changed indication. Error code: 0x%02X\n", err_code);
                        fflush(stdout);
                    }
                    break;
                }
                free(p_sys_attr_data);

            }
            break;

        // Service change indication acknolwedged
        case BLE_GATTS_EVT_SC_CONFIRM:
            printf("Service change acknowledged at time %llu\n", (GetTickCount64() - elapsedTimeStart));
            fflush(stdout);
            break;

		// Added just for info
		case BLE_GAP_EVT_CONN_PARAM_UPDATE:
			connParams = p_ble_evt->evt.gap_evt.params.conn_param_update.conn_params;
			printf("Connection parameters update at time %llu: min: %d max: %d supervision timeout: %d slave latency: %d \n",
				(GetTickCount64() - elapsedTimeStart),
				connParams.min_conn_interval, connParams.max_conn_interval, 
				connParams.conn_sup_timeout, connParams.slave_latency);
			break;

        case BLE_GAP_EVT_DISCONNECTED:
            printf("Disconnected at time %llu\n", (GetTickCount64() - elapsedTimeStart));
            fflush(stdout);
            // call twice; once to get the size of the data
            // create the buffer,
            // call a second time to load the data into the buffer
            sd_ble_gatts_sys_attr_get(m_adapter, p_ble_evt->evt.gap_evt.conn_handle, NULL, &saveDataLength, BLE_GATTS_SYS_ATTR_FLAG_SYS_SRVCS);
            if(saveDataLength > 0)
            { 
                saveDataBuffer = calloc(1, saveDataLength * sizeof(uint8_t));
                err_code = sd_ble_gatts_sys_attr_get(m_adapter, p_ble_evt->evt.gap_evt.conn_handle, saveDataBuffer, &saveDataLength, BLE_GATTS_SYS_ATTR_FLAG_SYS_SRVCS);
                if (err_code != NRF_SUCCESS)
                {
                    printf("Failed getting persistent sys attr info. Error code: 0x%02X\n", err_code);
                    fflush(stdout);
                }
            }
            m_connection_handle = BLE_CONN_HANDLE_INVALID;
            m_send_notifications = false;
            m_advertisement_timed_out = true;  // No restart
            //advertising_start();
            break;

        case BLE_GAP_EVT_TIMEOUT:
            printf("Advertisement timed out at time %llu\n", (GetTickCount64() - elapsedTimeStart));
            fflush(stdout);
            m_advertisement_timed_out = true;
            break;

        case BLE_GATTS_EVT_SYS_ATTR_MISSING:
            printf("Saved info being requested at time %llu\n", (GetTickCount64() - elapsedTimeStart));
            fflush(stdout);
            err_code = sd_ble_gatts_sys_attr_set(m_adapter, p_ble_evt->evt.gap_evt.conn_handle, saveDataBuffer,
                saveDataLength, BLE_GATTS_SYS_ATTR_FLAG_SYS_SRVCS);
            if (err_code != NRF_SUCCESS)
            {
                printf("Failed updating persistent sys attr info. Error code: 0x%02X\n", err_code);
                fflush(stdout);
            }
            break;

		// Added for pairing
		case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
			{
				ble_gap_sec_params_t sec_params = p_ble_evt->evt.gap_evt.params.sec_params_request.peer_params;
				printf("Peer requesting pairing at time %llu; Peer parameters:"
					"\n    bonding: %d"
					"\n    io_caps: %d"
					"\n    max key size: %d"
					"\n    min key size: %d"
					"\n    MITM supported: %d"
					"\n    Out of band supported: %d"
					"\n    secure connections (LESC) supported: %d"
					"\n    enable generation of keypress supported: %d"
					"\n    peer keys distributed: LTK %d  IRK %d  LNK from LTK %d  SIGN %d"
					"\n    own keys requested: LTK %d  IRK %d  LNK from LTK %d  SIGN %d\n",
					(GetTickCount64() - elapsedTimeStart),
					sec_params.bond, sec_params.io_caps, sec_params.max_key_size, sec_params.min_key_size,
					sec_params.mitm, sec_params.oob, sec_params.lesc, sec_params.keypress,
					sec_params.kdist_peer.enc, sec_params.kdist_peer.id, sec_params.kdist_peer.link, sec_params.kdist_peer.sign,
					sec_params.kdist_own.enc,  sec_params.kdist_own.id,  sec_params.kdist_own.link,  sec_params.kdist_own.sign);
				own_sec_params->bond = 1;
				own_sec_params->io_caps = BLE_GAP_IO_CAPS_NONE;
				own_sec_params->keypress = 0;
				own_sec_params->lesc = 0;
				own_sec_params->max_key_size = 16;
				own_sec_params->min_key_size = 8;
				own_sec_params->mitm = 0;
				own_sec_params->oob = 0;
				own_sec_params->kdist_peer.enc = 1; // sec_params.kdist_peer;
				own_sec_params->kdist_peer.id = 1;
				own_sec_params->kdist_peer.link = 0;	// This must be 0 or you will get an NRF_ERROR_NOT_SUPPORTED error
			    own_sec_params->kdist_peer.sign = 0;	// This must be 0 or you will get an NRF_ERROR_NOT_SUPPORTED error
				own_sec_params->kdist_own.enc = 1;
				own_sec_params->kdist_own.id = 1;
				own_sec_params->kdist_own.link = 0;		// This must be 0 or you will get an NRF_ERROR_NOT_SUPPORTED error
				own_sec_params->kdist_own.sign = 0;		// This must be 0 or you will get an NRF_ERROR_NOT_SUPPORTED error
				printf("Responding with own parameters at time %llu; Own parameters:"
					"\n    bonding: %d"
					"\n    io_caps: %d"
					"\n    max key size: %d"
					"\n    min key size: %d"
					"\n    MITM supported: %d"
					"\n    Out of band supported: %d"
					"\n    secure connections (LESC) supported: %d"
					"\n    enable generation of keypress supported: %d"
					"\n    peer keys requested: LTK %d  IRK %d  LNK from LTK %d  SIGN %d"
					"\n    own keys distributed: LTK %d  IRK %d  LNK from LTK %d  SIGN %d\n",
					(GetTickCount64() - elapsedTimeStart),
					own_sec_params->bond, own_sec_params->io_caps, own_sec_params->max_key_size, own_sec_params->min_key_size,
					own_sec_params->mitm, own_sec_params->lesc, own_sec_params->oob, own_sec_params->keypress,
					own_sec_params->kdist_peer.enc, own_sec_params->kdist_peer.id, own_sec_params->kdist_peer.link, own_sec_params->kdist_peer.sign,
					own_sec_params->kdist_own.enc,  own_sec_params->kdist_own.id,  own_sec_params->kdist_own.link,  own_sec_params->kdist_own.sign);
				uint32_t result = sd_ble_gap_sec_params_reply(adapter,
						p_ble_evt->evt.gap_evt.conn_handle,
						BLE_GAP_SEC_STATUS_SUCCESS,
						own_sec_params,
						keys);
				printf("Result of sending security params %d\n", result);
			}
			break;

		case BLE_GAP_EVT_CONN_SEC_UPDATE:
		{
			ble_gap_conn_sec_mode_t sec_mode = p_ble_evt->evt.gap_evt.params.conn_sec_update.conn_sec.sec_mode;
			printf("Security settings reported at time %llu:"
				"\n    security level: %d"
				"\n    security mode: %d"
				"\n    key length: %d\n", (GetTickCount64() - elapsedTimeStart),
				sec_mode.lv, sec_mode.sm, p_ble_evt->evt.gap_evt.params.conn_sec_update.conn_sec.encr_key_size);
            if (hasPaired)
            {
                m_is_encrytped = true;
                cccdSet = true;
            }
		}
		break;

		/* This event is signaled when pairing has completed. All the key information is
		 * placed in the 'keys' struct passed into the sd_ble_gap_sec_params_reply() method. 
		 */
		case BLE_GAP_EVT_AUTH_STATUS:
		{
			ble_gap_evt_auth_status_t authStatus = p_ble_evt->evt.gap_evt.params.auth_status;
			printf("Result of pairing %d at time %llu:"
				"\n    bonded: %d"
				"\n    peer keys: LTK %d  IRK %d  LNK from LTK %d  SIGN %d"
				"\n    own keys: LTK %d  IRK %d  LNK from LTK %d  SIGN %d\n",
				authStatus.auth_status, (GetTickCount64() - elapsedTimeStart),
				authStatus.bonded,
				authStatus.kdist_peer.enc, authStatus.kdist_peer.id, authStatus.kdist_peer.link, authStatus.kdist_peer.sign,
				authStatus.kdist_own.enc, authStatus.kdist_own.id, authStatus.kdist_own.link, authStatus.kdist_own.sign);

			byteToHex(keys->keys_peer.p_enc_key->enc_info.ltk, &byteString[0], ", ", keys->keys_peer.p_enc_key->enc_info.ltk_len);
			printf("Keys:\n    peer LTK %s", &byteString[0]);
			byteToHex(keys->keys_peer.p_id_key->id_info.irk, &byteString[0], ", ", 16);
			printf("\n    peer IRK %s", &byteString[0]);
			byteToHex(keys->keys_peer.p_id_key->id_addr_info.addr, &byteString[0], ", ", 6);
			printf("\n    peer Addrress %s", &byteString[0]);
			byteToHex(keys->keys_own.p_enc_key->enc_info.ltk, &byteString[0], ", ", keys->keys_own.p_enc_key->enc_info.ltk_len);
			printf("\n    own LTK  %s", &byteString[0]);
			byteToHex(keys->keys_own.p_id_key->id_info.irk, &byteString[0], ", ", 16);
			printf("\n    own IRK %s\n", &byteString[0]);
			hasPaired = true;
		}
		break;

		/* This event is signaled when encryption is being requested. Should only happen on a bonded reconnect. 
		 * Here the keys that are needed are those of the peripheral. The keys of the slave device are used */
		case BLE_GAP_EVT_SEC_INFO_REQUEST:
		{
			ble_gap_evt_sec_info_request_t sec_info = p_ble_evt->evt.gap_evt.params.sec_info_request;
			ble_gap_enc_info_t* enc_info = NULL;
			ble_gap_irk_t* irk_info = NULL;

			if (sec_info.enc_info) // Set if peer is asked for the encryption info
			{
				enc_info = &keys->keys_own.p_enc_key->enc_info; // Keys created by the peripheral are used
			}
			if (sec_info.id_info) // Set if peer is asking for the IRK
			{
				irk_info = &keys->keys_own.p_id_key->id_info;
			}
			uint32_t result = sd_ble_gap_sec_info_reply(adapter, p_ble_evt->evt.gap_evt.conn_handle,
				enc_info,
				irk_info,
				NULL);
			printf("Result of sending encryption params %d\n", result);
		}
		break;

		// Added to trap reads (for testing) - do not think it is necessary to trap writes
		case BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST:
		{
			printf("Read write authorized request signaled\n");
			ble_gatts_evt_rw_authorize_request_t auth = p_ble_evt->evt.gatts_evt.params.authorize_request;
			if (auth.type == BLE_GATTS_AUTHORIZE_TYPE_READ)
			{
				printf("Read authorized request at time %llu for characteristic %04X\n", (GetTickCount64() - elapsedTimeStart),
					auth.request.read.uuid.uuid);
				ble_gatts_rw_authorize_reply_params_t p_rw_authorize_reply_params;
				uint8_t systemId[8] = { 0xFE, 0xED, 0xAB, 0xEE, 0xDE, 0xAD, 0xBE, 0xEF }; // little endian order!
				uint8_t regCertDataList[22] = { 0, 2, 0, 0x12, 2, 1, 0, 8, 5, 0, 0, 1, 0, 2, 0x80, 6, 2, 2, 0, 2, 0x80, 0 };
				p_rw_authorize_reply_params.type = BLE_GATTS_AUTHORIZE_TYPE_READ;
				p_rw_authorize_reply_params.params.read.offset = 0;
				p_rw_authorize_reply_params.params.read.gatt_status = NRF_SUCCESS;
				p_rw_authorize_reply_params.params.read.update = 1;
				switch (auth.request.read.uuid.uuid)
				{
					case BTLE_DIS_SYSTEM_ID_CHAR:
						p_rw_authorize_reply_params.params.read.p_data = hrParams->disCharValues->systemId;
						p_rw_authorize_reply_params.params.read.len = 8;
						break;
					case BTLE_DIS_MODEL_NUMBER_CHAR:
						p_rw_authorize_reply_params.params.read.p_data = hrParams->disCharValues->modelNumber;
						p_rw_authorize_reply_params.params.read.len = (uint16_t)strlen(hrParams->disCharValues->modelNumber);
						break;
					case BTLE_DIS_SERIAL_NUMBER_CHAR:
						p_rw_authorize_reply_params.params.read.p_data = hrParams->disCharValues->serialNumber;
						p_rw_authorize_reply_params.params.read.len = (uint16_t)strlen(hrParams->disCharValues->serialNumber);
						break;
					case BTLE_DIS_FIRMWARE_REVISION_CHAR:
						p_rw_authorize_reply_params.params.read.p_data = hrParams->disCharValues->firmwareRevision;
						p_rw_authorize_reply_params.params.read.len = (uint16_t)strlen(hrParams->disCharValues->firmwareRevision);
						break;
					case BTLE_DIS_HARDWARE_REVISION_CHAR:
						p_rw_authorize_reply_params.params.read.p_data = hrParams->disCharValues->hardwareRevision;
						p_rw_authorize_reply_params.params.read.len = (uint16_t)strlen(hrParams->disCharValues->hardwareRevision);
						break;
					case BTLE_DIS_SOFTWARE_REVISION_CHAR:
						p_rw_authorize_reply_params.params.read.p_data = hrParams->disCharValues->softwareRevision;
						p_rw_authorize_reply_params.params.read.len = (uint16_t)strlen(hrParams->disCharValues->softwareRevision);
						break;
					case BTLE_DIS_MANUFACTURER_DEVICE_NAME_CHAR:
						p_rw_authorize_reply_params.params.read.p_data = hrParams->disCharValues->manufacturerName;
						p_rw_authorize_reply_params.params.read.len = (uint16_t)strlen(hrParams->disCharValues->manufacturerName);
						break;
					case BTLE_DIS_REG_CERT_DATA_LIST_CHAR:
						p_rw_authorize_reply_params.params.read.p_data = hrParams->disCharValues->regCertDataList;
						p_rw_authorize_reply_params.params.read.len = hrParams->disCharValues->regCertDataListLength;
						break;

					case BTLE_HRM_SENSOR_LOCATION:
						{
						    unsigned char location = 2;
							p_rw_authorize_reply_params.params.read.p_data = &location;
							p_rw_authorize_reply_params.params.read.len = 1;
						}
						break;
					default:
						p_rw_authorize_reply_params.params.read.update = 0;
				}
				uint32_t response = sd_ble_gatts_rw_authorize_reply(adapter, p_ble_evt->evt.gatts_evt.conn_handle, &p_rw_authorize_reply_params);
				printf("Read update reply response %u\n", response);
			}
			else if (auth.type == BLE_GATTS_AUTHORIZE_TYPE_WRITE) // Really dont need this to trap writes
			{
				printf("Write authorized request at time %llu\n", (GetTickCount64() - elapsedTimeStart));
				ble_gatts_rw_authorize_reply_params_t p_rw_authorize_reply_params;
				p_rw_authorize_reply_params.params.read.update = 1; // Must be set on a write
				sd_ble_gatts_rw_authorize_reply(adapter, p_ble_evt->evt.gatts_evt.conn_handle, &p_rw_authorize_reply_params);
			}
			else
			{
				printf("Received an authorized RW request at time %llu with type %d\n", (GetTickCount64() - elapsedTimeStart), auth.type);
			}
		}
		break;

        case BLE_GATTS_EVT_WRITE:
            if (p_ble_evt->evt.gatts_evt.params.write.handle ==
                    m_heart_rate_measurement_handle.cccd_handle)
            {
                uint8_t write_data = p_ble_evt->evt.gatts_evt.params.write.data[0];
                m_send_notifications = (write_data == BLE_GATT_HVX_NOTIFICATION);
                cccdSet = m_send_notifications;
				printf("Enabling CCCD with %d at time %llu\n", write_data, (GetTickCount64() - elapsedTimeStart));
            }
            else
            {
                uint8_t write_data = p_ble_evt->evt.gatts_evt.params.write.data[0];
                    printf("Client Enabling CCCD of characeteristic handle %d with value %d at time %llu\n",
                        p_ble_evt->evt.gatts_evt.params.write.handle, write_data, (GetTickCount64() - elapsedTimeStart));
            }
            break;


        case BLE_GATTS_EVT_EXCHANGE_MTU_REQUEST:
            err_code = sd_ble_gatts_exchange_mtu_reply(adapter, m_connection_handle,
                                                       GATT_MTU_SIZE_DEFAULT);

            if (err_code != NRF_SUCCESS)
            {
                printf("MTU exchange request reply failed. Error code: 0x%02X\n", err_code);
                fflush(stdout);
            }
            break;



        case BLE_GATTS_EVT_HVN_TX_COMPLETE:

#ifdef DEBUG
            printf("Successfully transmitted a heart rate reading.\n");
            fflush(stdout);
#endif
            break;

        default:
            printf("Received an un-handled event with ID %d at time %llu\n", p_ble_evt->header.evt_id, (GetTickCount64() - elapsedTimeStart));
            fflush(stdout);
            break;
        }
}

#pragma region // -------------- freeMemory
static void freeMemory(void)
{
	free(own_sec_params);
	free(keys->keys_own.p_enc_key);
	free(keys->keys_own.p_id_key);
	free(keys->keys_own.p_sign_key);
	free(keys->keys_own.p_pk);
	free(keys->keys_peer.p_enc_key);
	free(keys->keys_peer.p_id_key);
	free(keys->keys_peer.p_sign_key);
	free(keys->keys_peer.p_pk);
	free(keys);
}
#pragma endregion

/** Main */

/**@brief Heart Rate Monitor specialization
 * Creates a heart rate peripheral according to the data configured in the s_hrParams struct.
 * @param[in]   serialPort: the COM port used by the nRF dongle
 * @param[in]   hrParams: a pointer to the heart rate parameters stuct
 */
int runHeartRateMonitor(char * serialPort, s_hrParams *hrParamIn)
{
    uint32_t error_code;
    char *   serial_port = DEFAULT_UART_PORT_NAME;
    uint32_t baud_rate = DEFAULT_BAUD_RATE;
	int msmtCount = 0;
    ble_gap_addr_t addrStruct;

    setGlobals();

    hrParams = hrParamIn;

	own_sec_params = calloc(1, sizeof(ble_gap_sec_params_t));
	keys = calloc(1, sizeof(ble_gap_sec_keyset_t));

	keys->keys_own.p_enc_key = calloc(1, sizeof(ble_gap_enc_key_t));
	keys->keys_own.p_id_key = calloc(1, sizeof(ble_gap_id_key_t));
	keys->keys_own.p_sign_key = calloc(1, sizeof(ble_gap_sign_info_t));
	keys->keys_own.p_pk = calloc(1, sizeof(ble_gap_lesc_p256_pk_t));

	keys->keys_peer.p_enc_key = calloc(1, sizeof(ble_gap_enc_key_t));
	keys->keys_peer.p_id_key = calloc(1, sizeof(ble_gap_id_key_t));
	keys->keys_peer.p_sign_key = calloc(1, sizeof(ble_gap_sign_info_t));
	keys->keys_peer.p_pk = calloc(1, sizeof(ble_gap_lesc_p256_pk_t));

	loadKeysFromFile(hrParams->deviceName, keys, saveDataBuffer, &saveDataLength, &cccdSet);
    if (hrParams->invokeServiceChanged)
    {
        saveDataBuffer = NULL;
        saveDataLength = 0;
    }

    if (serialPort != NULL && strlen(serialPort) > 4)
    {
        serial_port = serialPort;
    }

    printf("Serial port used: %s\n", serial_port);
    printf("Baud rate used: %d\n", baud_rate);
    fflush(stdout);

    m_adapter =  adapter_init(serial_port, baud_rate);
    sd_rpc_log_handler_severity_filter_set(m_adapter, SD_RPC_LOG_INFO);
    error_code = sd_rpc_open(m_adapter, status_handler, ble_evt_dispatch, log_handler);

    if (error_code != NRF_SUCCESS)
    {
        printf("Failed to open nRF BLE Driver. Error code: 0x%02X\n", error_code);
        fflush(stdout);
		freeMemory();
        return error_code;
    }


    error_code = ble_cfg_set(m_config_id);

    if (error_code != NRF_SUCCESS)
    {
		freeMemory();
        return error_code;
    }

    error_code = ble_stack_init();

    if (error_code != NRF_SUCCESS)
    {
		freeMemory();
        return error_code;
    }

    ble_gap_addr_t currentAddrStruct;
    sd_ble_gap_addr_get(m_adapter, &currentAddrStruct);
    if (memcmp(currentAddrStruct.addr, hrParams->bluetoothAddress, 6) != 0)
    {
        addrStruct.addr_type = BLE_GAP_ADDR_TYPE_PUBLIC;
        memcpy(addrStruct.addr, hrParams->bluetoothAddress, 6);
        error_code = sd_ble_gap_addr_set(m_adapter, &addrStruct);

        if (error_code != NRF_SUCCESS)
        {
            printf("Could not set the Bluetooth Address\n");
            freeMemory();
            return error_code;
        }
    }
    else
    {
        printf("Bluetooth Address already set\n");
    }

    if (hrParams->disCharValues != NULL)
    {
        error_code = advertisement_data_set(m_adapter, &m_adv_handle,
            (uint16_t[]) {BTLE_HEART_RATE_SERVICE, BTLE_DEVICE_INFORMATION_SERVICE}, 2, NULL, hrParams->deviceName);

        if (error_code != NRF_SUCCESS)
        {
            freeMemory();
            return error_code;
        }

        error_code = createDeviceInformationService(m_adapter, &m_device_information_service_handle, hrParams->disCharValues);
        if (error_code != NRF_SUCCESS)
        {
            freeMemory();
            return error_code;
        }
    }
    else
    {
        error_code = advertisement_data_set(m_adapter, &m_adv_handle,
            (uint16_t[]) {BTLE_HEART_RATE_SERVICE}, 1, NULL, hrParams->deviceName);

        if (error_code != NRF_SUCCESS)
        {
            freeMemory();
            return error_code;
        }

    }

    error_code = createPrimaryService(m_adapter, &m_heart_rate_service_handle, BTLE_HEART_RATE_SERVICE);

    if (error_code != NRF_SUCCESS)
    {
		freeMemory();
        return error_code;
    }

	error_code = createStandardCharacteristic(m_adapter, m_heart_rate_service_handle,
		&m_heart_rate_measurement_handle,
		BTLE_HRM_MEASUREMENT_CHAR,
		true,
		false, // This will cause notify instead of indicate
		(ble_gap_conn_sec_mode_t) {1, 2},
		(ble_gap_conn_sec_mode_t) {0, 0},
		(ble_gap_conn_sec_mode_t) {0, 0});

	if (error_code != NRF_SUCCESS)
	{
		freeMemory();
		return error_code;
	}

	error_code = createStandardCharacteristic(m_adapter, m_heart_rate_service_handle,
		&m_sensor_location_handle,
		BTLE_HRM_SENSOR_LOCATION,
		false,
		false, // This will cause notify instead of indicate if hasCCCD is true
		(ble_gap_conn_sec_mode_t) {0, 0},
		(ble_gap_conn_sec_mode_t) {1, 1},
		(ble_gap_conn_sec_mode_t) {0, 0});

	if (error_code != NRF_SUCCESS)
	{
		freeMemory();
		return error_code;
	}

    error_code = advertising_start();

    if (error_code != NRF_SUCCESS)
    {
		freeMemory();
        return error_code;
    }

    while (!m_advertisement_timed_out)
    {
      //  printf("Looping: connection handle %d  notifications %d\n", m_connection_handle, m_send_notifications);
        if (m_connection_handle != BLE_CONN_HANDLE_INVALID && m_send_notifications)
        {
			if (msmtCount >= hrParams->numberOfMsmts)
			{
				sd_ble_gap_disconnect(m_adapter, m_connection_handle, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
				msmtCount = 0;
			}
			else
			{
                printf("Send mesurement %d\n", hrParams->hrMsmtChar[msmtCount]->heartRate);
				heart_rate_measurement_send(hrParams->hrMsmtChar[msmtCount]);
				msmtCount++;
			}
			Sleep(990);
        }

        Sleep(10);
    }

    error_code = sd_rpc_close(m_adapter);
	//if (hasPaired)
	{
        printf("Saving data to file at time %llu\n", (GetTickCount64() - elapsedTimeStart));
		saveKeysToFile(hrParams->deviceName, keys, saveDataBuffer, &saveDataLength, &cccdSet);
	}
    if (saveDataBuffer != NULL)
    {
        free(saveDataBuffer);
    }

    if (error_code != NRF_SUCCESS)
    {
        printf("Failed to close nRF BLE Driver. Error code: 0x%02X\n", error_code);
        fflush(stdout);
		freeMemory();
        return error_code;
    }

    printf("Closed at time %llu\n\n", (GetTickCount64() - elapsedTimeStart));
    fflush(stdout);

	freeMemory();
    return NRF_SUCCESS;
}
